/*
 *
 *
 */

#include "sim.h"

cmd_t cmdlist[] = {
  { "help", cmd_help, "Help" },
  { "quit", cmd_exit, "See exit" },
  { "exit", cmd_exit, "Quit" },
  { "dump_irq", (Function *)irq_dump, "Show IRQ statistics"},
  { "dump_alu", (Function *)dump_alu, "Show ALU registers"},
  { "dump_mac", (Function *)dump_mac, "Show MAC registers"},
  { "dump_shifter", (Function *)dump_shifter, "Show shifter registers"},
  { "dump_dags", (Function *)dump_dags, "Show DAG registers"},
  { "dump_pc", (Function *)pc_dump, "Show PC stack"},
  { "dump_status", (Function *)state_dump, "Show Status stack"},
  { "dump_dm", cmd_dumpdm, "Dump arg2 bytes at arg1 in Data Memory"},
  { "dump_pm", cmd_dumppm, "Dump arg2 bytes at arg1 in Program Memory"},
  { "dump_count", (Function *)count_dump, "Show Count stack"},
  { "dump_loop", (Function *)loop_dump, "Show Loop stack"},
  { "dump_breakpoints", (Function *)pcbp_dump, "List PC Breakpoints"},
  { "bdmasource", cmd_bdmasrc, "Load BDMA source from file"},
  { "boot", (Function *)bootdsp, "Perform External Bootstrap"},
  { "init", (Function *)dspinit, "Reset entire DSP"},
  { "step", cmd_step, "Step through execution at current PC"},
  { "jump", cmd_jump, "Sets a new PC (same as 'jump' instruction)"},
  { "call", cmd_call, "Calls a subroutine (same as 'call' instruction)"},
  { "setreg", cmd_setreg, "Set the value of a register"},
  { "pcbreak", cmd_pcbreak, "Set a PC breakpoint"},
  { "run", cmd_run, "Run til a Breakpoint"},
  { (char *)NULL, (Function *)NULL, (char *)NULL },
};

int initrl(void)
{
  rl_attempted_completion_function = (CPPFunction *)cmdcomplete;
  return 0;
}

int cmd_help(char *arg)
{
  int i;
  printf("\n");
  for (i=0;cmdlist[i].name;i++)
    printf("%15s\t\t%s\n", cmdlist[i].name, cmdlist[i].doc);
  printf("\n");
  return 0;
}

extern int stayalive;
int cmd_exit(char *arg)
{
  stayalive=0;
  return 0;
}

int cmd_bdmasrc(char *arg)
{
  struct stat st;
  if (stat(arg, &st)<0)
    {
      perror("bdmasrc: stat");
      return -1;
    }
  else if (st.st_size > BDMASRC_SIZE)
    printf("bdmasrc: data file too large, may truncate\n");
  
  return bdma_loadimage(arg);
}

int cmd_dumpdm(char *arg)
{
  int start=0,count=0;
  
  if ((sscanf(arg, "%d %d", &start, &count)<2) &&
      (sscanf(arg, "0x%x 0x%x", &start, &count)<2))
    {
      printf("dump_dm: syntax: dump_dm startaddr wordcount\n");
      return -1;
    }
  return dmdump(start, count);
}

int cmd_dumppm(char *arg)
{
  int start=0,count=0;
  
  if (sscanf(arg, "%d %d", &start, &count)<2)
    {
      printf("dump_pm: syntax: dump_pm startaddr wordcount\n");
      return -1;
    }
  return pmdump(start, count);
}

/*
 * Execute n instructions (where n defaults to 1).
 *
 */
int cmd_step(char *arg)
{
  int count=1;

  sscanf(arg, "%d", &count); /* if it fails, it will just fall to 1 */

  while(count)
    {
      idler();
      seq_fetch();
      count--;
    }
  return 0;
}

int cmd_run(char *arg)
{
  do
    {
      idler();
      seq_fetch();
    }
  while (!pcbp_check(curdsp.sequencer.pc));

  return 0;
}

int cmd_jump(char *arg)
{
  unsigned int addr = 0x000000;
  if (sscanf(arg, "0x%x", &addr)<1)
    {
      printf("cmd_jump: syntax: jump 0xaddress\n");
      return -1;
    }
  curdsp.sequencer.pc = addr;
  return 0;
}

int cmd_call(char *arg)
{
  unsigned int addr = 0x000000;
  if (sscanf(arg, "0x%x", &addr)<1)
    {
      printf("cmd_call: syntax: call 0xaddress\n");
      return -1;
    }
  pc_push();
  curdsp.sequencer.pc = addr;
  return 0;
}

int cmd_pcbreak(char *arg)
{
  unsigned int addr = 0x000000;
  if (sscanf(arg, "0x%x", &addr)<1)
    {
      printf("cmd_pcbreak: syntax: pcbreak 0xaddress\n");
      return -1;
    }
  pcbp_add(addr);
  return 0;
}

int cmd_setreg(char *arg)
{
  int reg, rgp;
  int reglen=0;
  int val=0;

  for (reglen=0;(arg[reglen]!=' ') && (arg[reglen]!=(char)NULL);reglen++)
    ;

  if (!findreg(arg, reglen, &reg, &rgp))
    {
      printf("cmd_setreg: unknown register\n");
      return -1;
    }
  
  if (sscanf(arg+reglen+1, "0x%x", &val)<1)
    {
      printf("cmd_setreg: bad value, must have 0x\n");
      return -1;
    }
  
  setregrgp(reg, rgp, val);

  return 0;
}

char *dupstr (char *s)
{
  char *r;

  r = xmalloc (strlen (s) + 1);
  strcpy (r, s);
  return (r);
}

char *cmdgenerator(char *text, int state)
{
  static int list_index, len;
  char *name;

  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }

  while ((name = cmdlist[list_index].name))
    {
      list_index++;
      if (strncmp (name, text, len) == 0)
        return (dupstr(name));
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

char **cmdcomplete(char *text, int start, int end)
{
  char **matches;
  
  matches = (char **)NULL;
  
  /* If this word is at the start of the line, then it is a command
     to complete.  Otherwise it is the name of a file in the current
     directory. */
  if (start == 0)
    matches = completion_matches(text, cmdgenerator);

  return (matches);
}

cmd_t *cmdfind(char *name)
{
  register int i;

  for (i = 0; cmdlist[i].name; i++)
    if (strcmp (name, cmdlist[i].name) == 0)
      return (&cmdlist[i]);

  return ((cmd_t *)NULL);
}


int cmdexec(char *line)
{
  register int i;
  cmd_t *command;
  char *word;

  /* Isolate the command word. */
  i = 0;
  while (line[i] && whitespace (line[i]))
    i++;
  word = line + i;

  while (line[i] && !whitespace (line[i]))
    i++;

  if (line[i])
    line[i++] = '\0';

  command = cmdfind(word);

  if (!command)
    {
      fprintf (stderr, "%s: invalid command\n", word);
      return (-1);
    }
  /* Get argument to command, if any. */
  while (whitespace (line[i]))
    i++;

  word = line + i;

  /* Call the function. */
  return ((*(command->func)) (word));
}

char *stripwhite (char *string)
{
  register char *s, *t;

  for (s = string; whitespace (*s); s++)
    ;

  if (*s == 0)
    return (s);

  t = s + strlen (s) - 1;
  while (t > s && whitespace (*t))
    t--;
  *++t = '\0';

  return s;
}

