/*
 * decode.c
 *
 * This is really where all the fun happens...
 *
 */

#include "sim.h"

typedef unsigned int (*opdecode_f)(unsigned int);

opdecode_f decodetable[] = {
  NULL,
  op_pmdmr,
  op_dmwim,
  op_dmrwim,
  op_dmrw,
  op_pmrw,
  op_loadregim,
  op_loadnondataregim,
  op_alumove,
  op_alu,
  op_jump,
  op_dountil,
  op_shiftdmrw,
  op_shiftpmrw,
  op_shiftmove,
  op_shiftim,
  op_shift,
  op_datamove,
  op_modectl,
  op_jumpindirect,
  op_return,
  op_modify,
  op_trap,
  op_divq,
  op_divs,
  op_satmr,
  op_stackctl,
  op_jumpfi,
  op_modifyfl,
  op_iorw,
  op_nop,
  op_idle,
  op_idlen
};


/*
 * 
 * Return the new PC if we commited a jump, else return 0 and the sequencer
 * will autoincrement for us.
 *
 */
unsigned int decode()
{
  unsigned int opcode;
  int type = 0;

  opcode = curdsp.sequencer.curinstr;
  
  if (opcode & BIT23) /* 1 */
    {
      if (opcode & BIT22) /* 11 -- type 1 */
	type = 11;
      else /* 10 */
	{
	  if (opcode & BIT21) /* 101 -- type 2 */
	    type = 2;
	  else /* 100 -- type 3 */
	    type = 3;
	}
    }
  else /* 0 */
    {
      if (opcode & BIT22) /* 01 */
	{
	  if (opcode & BIT21) /* 011 -- type 4 */
	    type = 4;
	  else /* 010 */
	    {
	      if (opcode & BIT20) /* 0101  -- type 5 */
		type = 5;
	      else /* 0100 -- type 6 */
		type = 6;
	    }
	}
      else  /* 00 */
	{
	  if (opcode & BIT21) /* 001 */
	    {
	      if (opcode & BIT20) /* 0011 -- type 7 */
		type = 7;
	      else /* 0010 */
		{
		  if (opcode & BIT19) /* 0010 1 -- type 8 */
		    type = 8;
		  else /* 0010 0 -- type 9 */
		    type = 9;
		}
	    }
	  else /* 000 */
	    {
	      if (opcode & BIT20) /* 0001 */
		{
		  if (opcode & BIT19) /* 00011 -- type 10 */
		    type = 10;
		  else /* 00010 */
		    {
		      if (opcode & BIT18) /* 000101 -- type 11 */
			type = 11;
		      else /* 000100 */
			{
			  if (opcode & BIT17) 
			    type = 12; /* type 12 */
			  else
			    {
			      if (opcode & BIT16)
				type = 13; /* type 13 */
			      else
				type = 14; /* type 14 */
			    }
			}
		    }
		}
	      else /* 0000 */
		{
		  if (opcode & BIT19) /* 0000 1 */
		    {
		      if (opcode & BIT18) /* 0000 11 */
			{
			  if (opcode & BIT17) /* 0000 111 */
			    {
			      if (opcode & BIT16) /* 0000 1111 */
				{
				  if (opcode & BIT15) 
				    ;
				  else /* type 15 */
				    type = 15;
				}
			      else /* 0000 1110 */
				{
				  if (opcode & BIT15) /* 0000 1110 1 */
				    ;
				  else /* type 16 */
				    type = 16;
				}
			    }
			  else /* 0000 110 */
			    {
			      if (opcode & BIT16) /* 0000 1101 -- type 17 */
				type = 17;
			      else /* 0000 1100 */
				type = 18; /* type 18 */
			    }
			}
		      else /* 0000 10 */
			{
			  if ((opcode >> 4) == 0xa001) /* rti */
			    type = 20; /* type 20 */
			  else if ((opcode >> 4) == 0xa000) /* rts */
			    type = 20; /* type 20 */
			  else if ((opcode >> 4) == 0x09080)
			    type = 21; /* type 21 */
			}
		    }
		  else /* 0000 0 */
		    {
		      if (opcode & BIT18) /* 0000 01 */
			{
			  if ((opcode >> 5) == 0x2000)
			    type = 26; /* type 26 */
			  else if (opcode & BIT17) /* 0000 011 */
			    {
			      if (opcode & BIT16) /* 0000 0111 */
				type = 23; /* type 23 */
			      else /* 0000 0110 */
				type = 24; /* type 24 */
			    }
			  else if (opcode == 0x050000)
			    type = 25; /* type 25 */
			}
		      else /* 000000 */
			{
			  if (opcode & BIT17) /* 0000 001 */
			    {
			      if (opcode & BIT16) /* 0000 0011 -- type 27 */
				type = 27;
			      else /* 0000 0010 -- type 28 */
				type = 28;
			    }
			  else /* 0000 000 */
			    {
			      if (opcode & BIT16) /* 0000 0001 -- type 29 */
				type = 29;
			      else /* 0000 0000 */
				{
				  if (opcode == 0x000000)
				    type = 30; /* type 30 */
				  else if ((opcode>>4) == 0x02800) 
				    {
				      if (opcode & 0xf)
					type = 32; /* type 31 */
				      else
					type = 31; /* type 31 */
				    }
				}
			    }
			}
		    }
		}
	    }
	}
    }

  if (type && (type <= 32))
    return (*(decodetable[type]))(opcode);
  
  fprintf(stderr, "decode: unknown instruction\n");
  return 0;
}
